# Type d'algorithme : 
# Auteur(s)         : Cedric MONDY
# Date              : 2018-05-25
# Version           : 1.0.1
# Interpreteur      : R version 3.5.0 (2018-04-23)
# Pre-requis        : Packages dplyr
# Fichiers lies     : ODInvertebres_params_normalisation.csv, ODInvertebres_params_traits.csv, ODInvertebres_params_transcodage.csv
# Commentaires      : 

# Copyright 2018 Cedric MONDY
# Ce programme est un logiciel libre; vous pouvez le redistribuer ou le modifier
# suivant les termes de la GNU General Public License telle que publiee par la
# Free Software Foundation; soit la version 3 de la licence, soit (a votre gre)
# toute version ulterieure.
# Ce programme est distribue dans l'espoir qu'il sera utile, mais SANS AUCUNE
# GARANTIE; sans meme la garantie tacite de QUALITE MARCHANDE ou d'ADEQUATION A
# UN BUT PARTICULIER. Consultez la GNU General Public License pour plus de
# details.
# Vous devez avoir recu une copie de la GNU General Public License en meme temps
# que ce programme; si ce n'est pas le cas, consultez
# <http://www.gnu.org/licenses>.

## VERSION ----

indic  <- "ODInvertebres"
vIndic <- "v1.0.1"

## CHARGEMENT DES PACKAGES ----

dependencies <- c("dplyr")

loadDependencies <- function(dependencies) {
    suppressAll <- function(expr) {
        suppressPackageStartupMessages(suppressWarnings(expr))
    }
    
    lapply(dependencies,
           function(x) 
           {
               suppressAll(library(x, character.only = TRUE))
           }
    )
    invisible()
}

loadDependencies(dependencies)

## DECLARATION DES FONCTIONS ----

## Fonction alternative a ifelse, plus simple
siNon <- function(test, yes, no) {
  if(test) {
    return(yes)
  } else {
    return(no)
  }
}

## Fonction permettant de tester si le tableau est un data frame
funDataFrame <- function(Table, tableau) {
  if (! "data.frame" %in% class(Table)) {
    return(list("ko",
                paste0("Le tableau ", 
                       tableau,
                       " n'est pas un data.frame"
                )
    )
    )
  } else {
    return("ok")
  }
}

## Fonction testant l'import des fichiers
funImport <- function(Table, empty = FALSE) {
  test <- siNon(empty, 
                is.null(Table), 
                is.null(Table) | nrow(Table) == 0)
  out <- siNon(empty,
               "Le fichier doit etre au bon format",
               "Le fichier doit etre au bon format et non vide")
    if (test) {
    return(list("ko", out))
  } else {
    return("ok")
  }
}

## Fonction permettant de tester la presence de champs obligatoires
funColonnes <- function(Table, cols) {
  # recupere le nom de l'objet passe a l'argument Table
  tableau <- deparse(substitute(Table))
  
  test <- funDataFrame(Table, tableau)
  
  if (test[[1]] == "ko") {
    return(test)
  } else {
    test <- which(! cols %in% colnames(Table))
    test <- test[sapply(test, length) > 0]
    
    if (length(test) > 0) {
      return(list("ko",
                  paste0("Les champs obligatoires suivants ", 
                         "ne sont pas presents : ", 
                         paste(cols[test], collapse = ", ")
                  )
      )
      )
    } else {
      return("ok")
    }
  }
  
}

## Fonction retournant un commentaire pour un test donné
funCommentaire <- function(test, message) {
  test <- test[sapply(test, length) > 0]
  
  if (length(test) > 0) {
    return(list("ko",
                lapply(1:length(test),
                       function(i) {
                         tibble(colonne = paste0("Colonne ", names(test)[i]),
                                ligne   = paste0("Ligne ", test[[i]] + 1),
                                message = message)
                       }
                ) %>%
                  bind_rows()
    )
    )
  } else {
    return("ok")
  }
}

## Fonction permettant de tester la presence de cellules vides
funVide <- function(Table) {
  tableau <- deparse(substitute(Table))
  
  test <- funDataFrame(Table, tableau)
  
  if (test[[1]] == "ko") {
    return(test)
  } else {
    
  testEmpty <- function(x) {
    return(which(is.na(x) | x == ""))
  }
  
  test <- lapply(Table, testEmpty)
  
  funCommentaire(test    = test,
                 message = "cellule vide"
  )
  }
}

## Fonction permettant de tester la presence d'espace
funEspace <- function(Table) {
  tableau <- deparse(substitute(Table))
  
  test <- funDataFrame(Table, tableau)
  
  if (test[1] == "ko") {
    return(test)
  } else {
    testSpace <- function(x) {
      which(grepl(" ", x))
    }
    
    test <- lapply(Table, testSpace)
    
    funCommentaire(test    = test, 
                   message = "cellule avec des caracteres 'espace'"
    )
  }
}

## Fonction permettant de tester si les valeurs sont numeriques
funNumerique <- function(Table, tableau = NULL) {
  if (is.null(tableau)) tableau <- deparse(substitute(Table))
  
  test <- funDataFrame(Table, tableau)
  
  if (test[1] == "ko") {
    return(test)
  } else {
    testNumeric <- function(x) {
      as.character(x)  %>%
        (function(i) {
          suppressWarnings(as.numeric(i))
        })           %>%
        is.na()      %>%
        which()
    }
    
    test <- lapply(Table, testNumeric)
    
    funCommentaire(test    = test,
                   message = "cellule avec valeur non numerique"
    )
  }
}

## Fonction permettant de tester si les nombres sont entiers
funEntier <- function(Table) {
  tableau <- deparse(substitute(Table))
  
  test <- funDataFrame(Table, tableau)
  
  if (test[1] == "ko") {
    return(test)
  } else {
    test <- funNumerique(Table, tableau = tableau)
    
    if (test[1] == "ko") {
      return(test)
    } else {
      testInteger <-   function(x) {
        as.numeric(x)                 %>%
          (function(x) abs(x - round(x)) > 
             .Machine$double.eps^0.5) %>%
          which()                     %>%
          return()
      }
      
      test <- lapply(Table, testInteger)
      
      funCommentaire(test    = test, 
                     message = "cellule avec valeur non entiere"
      )
    }
  }
}

## Fonction permettant de tester si les nombres sont positifs
funPositif <- function(Table) {
  tableau <- deparse(substitute(Table))
  
  test <- funDataFrame(Table, tableau)
  
  if (test[1] == "ko") {
    return(test)
  } else {
    test <- funNumerique(Table, tableau = tableau)
    
    if (test[1] == "ko") {
      return(test)
    } else {
      testPositive <- function(x) {
        return(which(x != abs(x)))
      }
      
      test <- lapply(Table, testPositive)
      
      funCommentaire(test    = test,
                     message = "cellule avec valeur non positive"
      )
    }
  }
}

## Fonction permettant de tester le format des dates
funDate <- function(Table, dateFormat) {
  tableau <- deparse(substitute(Table))
  
  test <- funDataFrame(Table, tableau)
  
  if (test[1] == "ko") {
    return(test)
  } else {
    
    testDate <- function(x, dateFormat) {
      suppressWarnings(as.Date(x, format = dateFormat)) %>%
        is.na()                                         %>%
        which()
    }
    
    test <- lapply(Table, testDate, dateFormat = dateFormat)
    
    funCommentaire(test    = test,
                   message = "cellule avec format de date non valide"
    )
  }
}

## Fonction permettant de tester des codes
funCodes <- function(Table, codes, codeType) {
  tableau <- deparse(substitute(Table))
  
  test <- funDataFrame(Table, tableau)
  
  if (test[1] == "ko") {
    return(test)
  } else {
    testCodes <- function(x, codes) {
      return(which(! x %in% codes))
    }
    
    test <- lapply(Table, testCodes, codes)
    
    funCommentaire(test    = test,
                   message = paste0("cellule avec code ", 
                                    codeType, " non valide"
                   )
    )
  }
}

## Fonction permettant de changer les valeurs de sortie/verif en fonction des
## resultats d'un test
funTest <- function(test, result) {
  result$sortie <- siNon(test[[1]] == "ko",
                         bind_rows(result$sortie,
                                   test[[2]]),
                         result$sortie)
  result$verif <- siNon(test[1] == "ko", "ko", result$verif)
  
  return(result)
}

## Fonction permettant de nettoyer les sorties
funSortie <- function(sortie) {
  msgLevels <- c("vide", "espace", 
                 "numerique", "entiere", "positive",
                 "date", "code")
  
  messages <- tibble(message = unique(sortie$message),
                     level   = NA)
  
  for (i in 1:length(msgLevels)) {
    i2 <- which(grepl(pattern = msgLevels[i], x = messages$message))
    
    messages$level[i2] <- i
  }
  
  messages <- arrange(messages, level) %>%
    mutate(message = factor(message, levels = message))
  
  mutate(sortie,
         message = factor(message, 
                          levels = messages$message)) %>%
    group_by(colonne, ligne)                          %>%
    summarise(message = message[min(as.numeric(message))])
}

## Fonction permettant d'ecrire le fichier de sortie
funResult		<- function ()
{
  # determination du temps de calcul
  heure_fin       <- Sys.time()
  heure_dif       <- heure_fin - heure_debut
  temps_execution <- paste(round(heure_dif,2), 
                           attr(heure_dif,"units"),
                           sep="")
  
  # creation du bandeau d'information
  etiquette <- paste(indic, vIndic, Sys.Date(),
                     "Temps d'execution :", temps_execution)
  
  # information de validation ou non des donnees d'entree
  print(valid)
  
  # sortie du bandeau d'information
  write(etiquette, file = "", sep = "")
  
  # sortie du rapport d'erreur de l'indice
  write("Fichier d'entree: ", 
        file = "", sep = "")
  write.table(sortie, sep = ";", col.names = FALSE,
              row.names = FALSE, quote = FALSE)

    # creation du fichier de sortie avec bandeau d'information
  outputFile <- paste0(indic, "_", vIndic, "_rapport_erreur.csv")
  
  cat(etiquette, file = outputFile, append = FALSE)
  cat("\n"     , file = outputFile, append = TRUE)
  
  # completion du fichier de sortie avec les erreurs
  cat("Fichier d'entree:\n", 
      file = outputFile, append = TRUE)
  write.table(sortie, file = outputFile, sep = ";",
              col.names = FALSE, row.names = FALSE,
              quote = FALSE, append = TRUE)
  
}# fin de la fonction funResult


## INITIALISATION DU TRAITEMENT ----

# Ne pas afficher les messages d'avis
options(warn = -1, show.error.messages = FALSE)

# Recuperation des fichiers d'entree
File <- "ODInvertebres_entree_op100.txt"

# Initialisation de l'heure
heure_debut <- Sys.time()

### IMPORT DES FICHIERS ----

# Import du fichier d'entree
data_entree <- read.table(File, header = TRUE, sep = "\t", as.is = TRUE)

## VALIDATION DES DONNEES ----

result <- list(sortie = tibble(colonne = "", ligne = "", message = "")[0,],
               verif  = "ok")

# Verifie si le fichier a bien ete importe et contient des donnees
test1 <- funImport(data_entree)
result <- funTest(test1, result)

if (result$verif == "ok") {
  # Verifie si les colonnes obligatoires sont bien presentes
  test2 <- funColonnes(data_entree,
                       cols = c("CODE_OPERATION", "CODE_TAXON",
                                "RESULTAT", "CODE_REMARQUE"))
  result <- funTest(test2, result)
  
  if (result$verif == "ok") {
    # Verifie que les cellules sont non vides
    test3 <- funVide(select(data_entree, 
                            CODE_OPERATION,
                            CODE_TAXON, RESULTAT))
    result <- funTest(test3, result)
    
    # Verifie l'absence d'espaces
    test4 <- funEspace(select(data_entree, 
                              CODE_TAXON, RESULTAT))
    result <- funTest(test4, result)
    
    # Verifie que les resultats sont des valeurs numeriques positives
    test5 <- funPositif(select(data_entree, RESULTAT))
    result <- funTest(test5, result)
    
    # Verifie que les codes remarques sont corrects
    test6 <- funCodes(select(data_entree, CODE_REMARQUE), 
                      codes = c(1, 4), codeType = "remarque")
    result <- funTest(test6, result)
    
  }
}

# Parametre de succes/echec de la validation
valid <- sapply(list(result),
                function(x) {
                  x$verif == "ko"
                }) %>%
  (function(x) siNon(any(x), "ko", "ok"))

sortie <- siNon(result$verif == "ok",
                "", 
                funSortie(result$sortie))

# Sortie du rapport d'erreur
funResult()
